# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


import re
from binascii import unhexlify
from typing import Optional

from amazon_transcribe import AWSCRTEventLoop
from amazon_transcribe.auth import AwsCrtCredentialResolver, CredentialResolver
from amazon_transcribe.endpoints import (
    BaseEndpointResolver,
    _TranscribeRegionEndpointResolver,
)
from amazon_transcribe.eventstream import EventStreamMessageSerializer
from amazon_transcribe.eventstream import EventSigner
from amazon_transcribe.httpsession import AwsCrtHttpSessionManager
from amazon_transcribe.model import (
    AudioStream,
    StartStreamTranscriptionRequest,
    StartStreamTranscriptionEventStream,
)
from amazon_transcribe.serialize import (
    AudioEventSerializer,
    TranscribeStreamingSerializer,
)
from amazon_transcribe.deserialize import TranscribeStreamingResponseParser
from amazon_transcribe.signer import SigV4RequestSigner


class TranscribeStreamingClient:
    """High level client for orchestrating setup and transmission of audio
    streams to Amazon TranscribeStreaming service.

    :param region: An AWS region to use for Amazon Transcribe (e.g. us-east-2)
    :param endpoint_resolver: Optional resolver for client endpoints.
    :param credential_resolver: Optional credential resolver for client.
    """

    def __init__(
        self,
        *,
        region: str,
        endpoint_resolver: Optional[BaseEndpointResolver] = None,
        credential_resolver: Optional[CredentialResolver] = None,
    ):
        if endpoint_resolver is None:
            endpoint_resolver = _TranscribeRegionEndpointResolver()
        self._endpoint_resolver = endpoint_resolver
        self.service_name = "transcribe"
        self.region = region
        self._event_signer = EventSigner(self.service_name, self.region)
        self._eventloop = AWSCRTEventLoop().bootstrap
        if credential_resolver is None:
            credential_resolver = AwsCrtCredentialResolver(self._eventloop)
        self._credential_resolver = credential_resolver
        self._response_parser = TranscribeStreamingResponseParser()
        self._serializer = TranscribeStreamingSerializer()

    async def start_stream_transcription(
        self,
        *,
        language_code: str,
        media_sample_rate_hz: int,
        media_encoding: str,
        vocabulary_name: Optional[str] = None,
        session_id: Optional[str] = None,
        vocab_filter_method: Optional[str] = None,
        vocab_filter_name: Optional[str] = None,
        show_speaker_label: Optional[bool] = None,
        enable_channel_identification: Optional[bool] = None,
        number_of_channels: Optional[int] = None,
    ) -> StartStreamTranscriptionEventStream:
        """Coordinate transcription settings and start stream.

        Pay careful attention to language_code and media_sample_rate_hz configurations.
        Incorrect setups may lead to streams hanging indefinitely.
        More info on constraints can be found here:
        https://docs.aws.amazon.com/transcribe/latest/dg/streaming.html

        The service processes audio chunks for this operation in realtime, if
        the audio chunks provided are not coming from a realtime source (e.g. a
        pre-recorded file) the audio chunks should be rate limited to match the
        appropriate realtime bitrate for the stream. Failure to send the audio
        chunks in realtime can lead to signing issues for audio streams longer
        than 5 minutes.

        :param language_code:
            Indicates the source language used in the input audio stream.
        :param media_sample_rate_hz:
            The sample rate, in Hertz, of the input audio. We suggest that you
            use 8000 Hz for low quality audio and 16000 Hz for high quality audio.
        :param media_encoding:
            The encoding used for the input audio.
        :param vocabulary_name:
            The name of the vocabulary to use when processing the transcription job.
        :param session_id:
            A identifier for the transcription session. Use this parameter when you
            want to retry a session. If you don't provide a session ID,
            Amazon Transcribe will generate one for you and return it in the response.
        :param vocab_filter_method:
            The manner in which you use your vocabulary filter to filter words in
            your transcript. See Transcribe Streaming API docs for more info.
        :param vocab_filter_name:
            The name of the vocabulary filter you've created that is unique to
            your AWS account. Provide the name in this field to successfully
            use it in a stream.
        :param show_speaker_label:
            When true, enables speaker identification in your real-time stream.
        :param enable_channel_identification:
            When true, instructs Amazon Transcribe to process each audio channel
            separately and then merge the transcription output of each channel
            into a single transcription. Amazon Transcribe also produces a
            transcription of each item. An item includes the start time, end time,
            and any alternative transcriptions. You can't set both ShowSpeakerLabel
            and EnableChannelIdentification in the same request. If you set both,
            your request returns a BadRequestException.
        :param number_of_channels:
            The number of channels that are in your audio stream.
        """
        transcribe_streaming_request = StartStreamTranscriptionRequest(
            language_code,
            media_sample_rate_hz,
            media_encoding,
            vocabulary_name,
            session_id,
            vocab_filter_method,
            vocab_filter_name,
            show_speaker_label,
            enable_channel_identification,
            number_of_channels,
        )
        endpoint = await self._endpoint_resolver.resolve(self.region)

        request = self._serializer.serialize_start_stream_transcription_request(
            endpoint=endpoint, request_shape=transcribe_streaming_request,
        ).prepare()

        creds = await self._credential_resolver.get_credentials()
        signer = SigV4RequestSigner("transcribe", self.region)
        signed_request = signer.sign(request, creds)

        session = AwsCrtHttpSessionManager(self._eventloop)

        response = await session.make_request(
            signed_request.uri,
            method=signed_request.method,
            headers=signed_request.headers.as_list(),
            body=signed_request.body,
        )
        resolved_response = await response.resolve_response()

        status_code = resolved_response.status_code
        if status_code >= 400:
            # We need to close before we can consume the body or this will hang
            signed_request.body.close()
            body_bytes = await response.consume_body()
            raise self._response_parser.parse_exception(resolved_response, body_bytes)
        elif status_code != 200:
            raise RuntimeError("Unexpected status code encountered: %s" % status_code)

        parsed_response = self._response_parser.parse_start_stream_transcription_response(
            resolved_response, response,
        )

        # The audio stream is returned as output because it requires
        # the signature from the initial HTTP request to be useable
        audio_stream = self._create_audio_stream(signed_request)
        return StartStreamTranscriptionEventStream(audio_stream, parsed_response)

    def _create_audio_stream(self, signed_request):
        initial_signature = self._extract_signature(signed_request)
        return AudioStream(
            input_stream=signed_request.body,
            event_serializer=AudioEventSerializer(),
            eventstream_serializer=EventStreamMessageSerializer(),
            event_signer=self._event_signer,
            initial_signature=initial_signature,
            credential_resolver=self._credential_resolver,
        )

    def _extract_signature(self, signed_request):
        auth = signed_request.headers.get("Authorization", "")
        auth = re.split("Signature=", auth)[-1]
        return unhexlify(auth)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


import asyncio
from typing import Optional

from awscrt.auth import AwsCredentialsProvider


class Credentials:
    def __init__(
        self,
        access_key_id: str,
        secret_access_key: str,
        session_token: Optional[str] = None,
    ):
        self.access_key_id = access_key_id
        self.secret_access_key = secret_access_key
        self.session_token = session_token


class CredentialResolver:
    async def get_credentials(self) -> Optional[Credentials]:
        raise NotImplementedError("get_credentials")


class AwsCrtCredentialResolver(CredentialResolver):
    """Default Credential resolution chain provided by CRT."""

    def __init__(self, eventloop):
        self._crt_resolver = AwsCredentialsProvider.new_default_chain(eventloop)

    async def get_credentials(self) -> Optional[Credentials]:
        credentials = await asyncio.wrap_future(self._crt_resolver.get_credentials())
        return credentials


class StaticCredentialResolver(Credentials, CredentialResolver):
    """Hardcoded credentials provided from a static source."""

    async def get_credentials(self) -> Optional[Credentials]:
        return self



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


class SDKError(Exception):
    """Base Error for Amazon Transcribe Streaming SDK"""


class HTTPException(SDKError):
    """Base error for HTTP related exceptions"""


class ServiceException(SDKError):
    """Errors returned by the service"""


class UnknownServiceException(ServiceException):
    def __init__(self, status_code, error_code, message):
        self.status_code: int = status_code
        self.error_code: int = error_code
        self.message: str = message


class BadRequestException(ServiceException):
    def __init__(self, message):
        self.message: str = message
        self.status_code: int = 400


class ConflictException(ServiceException):
    def __init__(self, message):
        self.message: str = message
        self.status_code = 409


class InternalFailureException(ServiceException):
    def __init__(self, message):
        self.message: str = message
        self.status_code: int = 500


class LimitExceededException(ServiceException):
    def __init__(self, message):
        self.message: str = message
        self.error: int = 429


class ServiceUnavailableException(ServiceException):
    def __init__(self, message):
        self.message: str = message
        self.status_code: int = 503


class SerializationException(ServiceException):
    def __init__(self, message):
        self.message: str = message
        self.status_code: int = 400


class SerializerException(SDKError):
    """Encountered an issue when seralizing a request or event"""


class ValidationException(SDKError):
    """Encountered an issue validating a given value"""


class CredentialsException(SDKError):
    """Encountered an issue while resolving or using credentials"""



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


from typing import Optional

from awscrt.http import HttpHeaders, HttpRequest
from awscrt.auth import (
    AwsCredentialsProvider,
    AwsSigningAlgorithm,
    AwsSigningConfig,
    AwsSignatureType,
    AwsSignedBodyValue,
    AwsSignedBodyHeaderType,
    aws_sign_request,
)
from amazon_transcribe.auth import Credentials
from amazon_transcribe.request import PreparedRequest, HeadersDict
from amazon_transcribe.exceptions import CredentialsException


class RequestSigner:
    """General implementation for Request signing"""

    def __init__(self, service_name, region, algorithm=0, signature_type=0):
        self.service_name: str = service_name
        self.region: str = region
        self.algorithm: int = algorithm
        self.signature_type: int = signature_type

    def sign(
        self, request: PreparedRequest, credentials: Optional[Credentials]
    ) -> PreparedRequest:
        if credentials is None:
            raise CredentialsException("Failed to resolve credentials")
        alg = AwsSigningAlgorithm(self.algorithm)
        sig_type = AwsSignatureType(self.signature_type)

        credential_provider = AwsCredentialsProvider.new_static(
            credentials.access_key_id,
            credentials.secret_access_key,
            credentials.session_token,
        )

        config = AwsSigningConfig(
            algorithm=alg,
            signature_type=sig_type,
            credentials_provider=credential_provider,
            region=self.region,
            service=self.service_name,
            signed_body_value=AwsSignedBodyValue.EMPTY_SHA256,
            signed_body_header_type=AwsSignedBodyHeaderType.NONE,
        )
        crt_request = _convert_request(request)
        signed_request = aws_sign_request(crt_request, config).result()
        request.headers = HeadersDict(dict(signed_request.headers))

        return request


class SigV4RequestSigner(RequestSigner):
    def __init__(self, service_name, region):
        super().__init__(service_name, region)
        self.algorithm: int = AwsSigningAlgorithm.V4
        self.signature_type: int = AwsSignatureType.HTTP_REQUEST_HEADERS


def _convert_request(request: PreparedRequest) -> HttpRequest:
    return HttpRequest(
        method=request.method,
        path=request.path,
        headers=HttpHeaders(request.headers.as_list()),
        body_stream=request.body,
    )



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


from collections.abc import MutableMapping
from io import BufferedIOBase, BytesIO
from typing import Dict, List, Tuple, Union

from amazon_transcribe.exceptions import ValidationException


LIST_TYPE = Union[Tuple[str, ...], List[str]]
HEADER_VALUE_TYPE = Union[str, LIST_TYPE]
BODY_TYPE = Union[BytesIO, BufferedIOBase]


class _HeaderKey:
    def __init__(self, key: str):
        self._key = key
        self._lower = key.lower()

    def __hash__(self):
        return hash(self._lower)

    def __eq__(self, other):
        return isinstance(other, _HeaderKey) and self._lower == other._lower

    def __str__(self):
        return self._key

    def __repr__(self):
        return repr(self._key)


class HeadersDict(MutableMapping):
    """A case-insenseitive dictionary to represent HTTP headers."""

    def __init__(self, *args, **kwargs):
        self._dict: Dict = {}
        self.update(*args, **kwargs)

    def __setitem__(self, key: str, value: HEADER_VALUE_TYPE):
        key, value = self._validate_header(key, value)
        self._dict[_HeaderKey(key)] = value

    def __getitem__(self, key: str):
        return self._dict[_HeaderKey(key)]

    def __delitem__(self, key: str):
        del self._dict[_HeaderKey(key)]

    def __iter__(self):
        return (str(key) for key in self._dict)

    def __len__(self):
        return len(self._dict)

    def __repr__(self):
        return repr(self._dict)

    def copy(self) -> "HeadersDict":
        return HeadersDict(self.items())

    def as_list(self) -> List[Tuple[str, str]]:
        return [(str(k), v) for k, v in self._dict.items() if v is not None]

    def _validate_str(self, string: str) -> str:
        if string is None:
            return string
        # newline characters are prohibited in headers
        for seq in ("\r\n", "\r", "\n"):
            string = string.replace(seq, "")
        return string.strip(" ")

    def _validate_header_list(
        self, key: str, values: LIST_TYPE
    ) -> Tuple[str, HEADER_VALUE_TYPE]:
        value_list = [self._validate_str(v) for v in values if v is not None]
        return self._validate_str(key), ";".join(value_list)

    def _validate_header(
        self, key: str, value: HEADER_VALUE_TYPE
    ) -> Tuple[str, HEADER_VALUE_TYPE]:
        if key is None:
            raise ValidationException("Unexpected key (None) was provided in headers")
        if isinstance(value, (tuple, list)):
            return self._validate_header_list(key, value)
        elif not isinstance(value, str) and value is not None:
            value = str(value)
        return self._validate_str(key), self._validate_str(value)


class Request:
    def __init__(
        self, endpoint, path="/", method="GET", headers=None, body=None, params=None,
    ):
        self.endpoint: str = endpoint
        self.path: str = path
        self.method: str = method
        self.headers: Dict = headers if headers is not None else {}
        self.params: Dict = params if params is not None else {}
        self.body = body

    def prepare(self):
        method: str = self.prepare_method()
        query_str: str = self.prepare_params()
        headers: HeadersDict = self.prepare_headers()
        body: BODY_TYPE = self.prepare_body()
        return PreparedRequest(
            self.endpoint, self.path, method, headers, body, query_str
        )

    def prepare_method(self) -> str:
        return self.method.upper()

    def prepare_params(self) -> str:
        """Converts dictionary of params into query string"""
        query_list = []
        for k, v in self.params.items():
            if v is None:
                # empty values should just apply the key
                # e.g. foo=None, bar=baz -> foo&bar=baz
                query_list.append(k)
            else:
                query_list.append(f"{k}={v}")
        return "&".join(query_list)

    def prepare_headers(self) -> HeadersDict:
        prepared_headers = HeadersDict()
        prepared_headers.update(self.headers)
        return prepared_headers

    def prepare_body(self) -> BODY_TYPE:
        body = self.body
        if body is None:
            return BytesIO(b"")
        elif isinstance(body, str):
            return BytesIO(body.encode("utf-8"))
        elif isinstance(body, bytes):
            return BytesIO(body)
        elif not isinstance(body, BufferedIOBase):
            type_ = type(body)
            raise ValidationException(
                f"Body provided is an unexpected type ({type_}). Request was "
                f"expecting bytes, str, or file-like body."
            )

        return body


class PreparedRequest:
    def __init__(self, endpoint, path, method, headers, body, query_str):
        self.endpoint: str = endpoint
        self.path: str = path
        self.method: str = method
        self.headers: HeadersDict = headers
        self.body: BytesIO = body
        self.query: str = query_str

    @property
    def uri(self) -> str:
        endpoint = self.endpoint.rstrip("/")
        path = self.path.lstrip("/")
        output_uri = "/".join([endpoint, path])
        if self.query:
            output_uri = "?".join([output_uri, self.query])
        return output_uri



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


from urllib.parse import urlsplit
from typing import Dict

from amazon_transcribe import __version__ as version
from amazon_transcribe.exceptions import ValidationException


def _add_required_headers(endpoint: str, headers: Dict[str, str]):
    urlparts = urlsplit(endpoint)
    if not urlparts.hostname:
        raise ValidationException(
            "Unexpected endpoint ({endpoint}) provided to serializer"
        )
    headers.update(
        {
            "user-agent": f"transcribe-streaming-sdk-{version}",
            "host": urlparts.hostname,
        }
    )


def ensure_boolean(val):
    if isinstance(val, bool):
        return val
    else:
        return val.lower() == "true"



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from typing import Optional, List

from amazon_transcribe.eventstream import BaseEvent, BaseStream, EventStream


class Alternative:
    """A list of possible transcriptions for the audio.

    :param transcript: The text that was transcribed from the audio.
    :param items: One or more alternative interpretations of the input audio.
    """

    def __init__(self, transcript, items):
        self.transcript: str = transcript
        self.items: List[Item] = items


class AudioEvent(BaseEvent):
    """Provides a wrapper for the audio chunks that you are sending.

    :param audio_chunk:
        A blob of audio from your application. You audio stream
        consists of one or more audio events.
    """

    def __init__(self, audio_chunk: Optional[bytes]):
        if audio_chunk is None:
            audio_chunk = b""
        super().__init__(payload=audio_chunk)

    @property
    def audio_chunk(self):
        return self.payload


class AudioStream(BaseStream):
    """Input audio stream for transcription stream request.

    This should never be instantiated by the end user. It will be returned
    from the client within a relevant wrapper object.
    """

    async def send_audio_event(self, audio_chunk: Optional[bytes]):
        """Enqueue audio bytes to be sent for transcription.

        :param audio_chunk: byte-string chunk of audio input.
        """
        audio_event = AudioEvent(audio_chunk)
        await super().send_event(audio_event)


class Item:
    """A word or phrase transcribed from the input audio.

    :param start_time:
        The offset from the beginning of the audio stream to the beginning
        of the audio that resulted in the item.

    :param end_time:
        The offset from the beginning of the audio stream to the end of
        the audio that resulted in the item.

    :param item_type: The type of the item.

    :param content:
        The word or punctuation that was recognized in the input audio.

    :param vocabulary_filter_match:
        Indicates whether a word in the item matches a word in the vocabulary
        filter you've chosen for your real-time stream. If True then a word
        in the item matches your vocabulary filter.

    :param speaker:
        If speaker identification is enabled, shows the speakers identified
        in the real-time stream.
    """

    def __init__(
        self,
        start_time: Optional[float] = None,
        end_time: Optional[float] = None,
        item_type: Optional[str] = None,
        content: Optional[str] = None,
        vocabulary_filter_match: Optional[bool] = None,
        speaker: Optional[str] = None,
    ):
        self.start_time = start_time
        self.end_time = end_time
        self.item_type = item_type
        self.content = content
        self.vocabulary_filter_match = vocabulary_filter_match
        self.speaker = speaker


class Result:
    """The result of transcribing a portion of the input audio stream.

    :param result_id: A unique identifier for the result.

    :param start_time:
        The offset in seconds from the beginning of the audio stream to the
        beginning of the result.

    :param end_time:
        The offset in seconds from the beginning of the audio stream to the
        end of the result.

    :param is_partial:
        Amazon Transcribe divides the incoming audio stream into segments at
        natural points in the audio. Transcription results are returned based
        on these segments. True indicates that Amazon Transcribe has additional
        transcription data to send, False to indicate that this is the last
        transcription result for the segment.

    :param alternatives:
        A list of possible transcriptions for the audio. Each alternative
        typically contains one Item that contains the result of the transcription.

    :param channel_id:
        When channel identification is enabled, Amazon Transcribe transcribes
        the speech from each audio channel separately. You can use ChannelId
        to retrieve the transcription results for a single channel in your
        audio stream.
    """

    def __init__(
        self,
        result_id: Optional[str] = None,
        start_time: Optional[float] = None,
        end_time: Optional[float] = None,
        is_partial: Optional[bool] = None,
        alternatives: Optional[List[Alternative]] = None,
        channel_id: Optional[str] = None,
    ):
        self.result_id = result_id
        self.start_time = start_time
        self.end_time = end_time
        self.is_partial = is_partial
        self.alternatives = alternatives
        self.channel_id = channel_id


class StartStreamTranscriptionRequest:
    """Transcription Request

    :param language_code:
        Indicates the source language used in the input audio stream.

    :param media_sample_rate_hz:
        The sample rate, in Hertz, of the input audio. We suggest that you
        use 8000 Hz for low quality audio and 16000 Hz for high quality audio.

    :param media_encoding:
        The encoding used for the input audio.

    :param vocabulary_name:
        The name of the vocabulary to use when processing the transcription job.

    :param session_id:
        A identifier for the transcription session. Use this parameter when you
        want to retry a session. If you don't provide a session ID,
        Amazon Transcribe will generate one for you and return it in the response.

    :param vocab_filter_method:
        The manner in which you use your vocabulary filter to filter words in
        your transcript.

    :param vocab_filter_name:
        The name of the vocabulary filter you've created that is unique to
        your AWS account. Provide the name in this field to successfully
        use it in a stream.

    :param show_speaker_label:
        When true, enables speaker identification in your real-time stream.

    :param enable_channel_identification:
        When true, instructs Amazon Transcribe to process each audio channel
        separately and then merge the transcription output of each channel
        into a single transcription. Amazon Transcribe also produces a
        transcription of each item. An item includes the start time, end time,
        and any alternative transcriptions. You can't set both ShowSpeakerLabel
        and EnableChannelIdentification in the same request. If you set both,
        your request returns a BadRequestException.

    :param number_of_channels:
        The number of channels that are in your audio stream.
    """

    def __init__(
        self,
        language_code=None,
        media_sample_rate_hz=None,
        media_encoding=None,
        vocabulary_name=None,
        session_id=None,
        vocab_filter_method=None,
        vocab_filter_name=None,
        show_speaker_label=None,
        enable_channel_identification=None,
        number_of_channels=None,
    ):

        self.language_code: Optional[str] = language_code
        self.media_sample_rate_hz: Optional[int] = media_sample_rate_hz
        self.media_encoding: Optional[str] = media_encoding
        self.vocabulary_name: Optional[str] = vocabulary_name
        self.session_id: Optional[str] = session_id
        self.vocab_filter_method: Optional[str] = vocab_filter_method
        self.vocab_filter_name: Optional[str] = vocab_filter_name
        self.show_speaker_label: Optional[bool] = show_speaker_label
        self.enable_channel_identification: Optional[
            bool
        ] = enable_channel_identification
        self.number_of_channels: Optional[int] = number_of_channels


class StartStreamTranscriptionResponse:
    """Transcription Response

    :param transcript_result_stream:
        Represents the stream of transcription events from
        Amazon Transcribe to your application.

    :param request_id: An identifier for the streaming transcription.

    :param language_code:
        Indicates the source language used in the input audio stream.

    :param media_sample_rate_hz:
        The sample rate, in Hertz, of the input audio. We suggest that you
        use 8000 Hz for low quality audio and 16000 Hz for high quality audio.

    :param media_encoding:
        The encoding used for the input audio.

    :param session_id:
        A identifier for the transcription session. Use this parameter when you
        want to retry a session. If you don't provide a session ID,
        Amazon Transcribe will generate one for you and return it in the response.

    :param vocab_filter_name:
        The name of the vocabulary filter used in your real-time stream.

    :param vocab_filter_method:
        The manner in which you use your vocabulary filter to filter words in
        your transcript.

    :param show_speaker_label:
        Shows whether speaker identification was enabled in the stream.

    :param enable_channel_identification:
        Shows whether channel identification has been enabled in the stream.

    :param number_of_channels:
        The number of channels identified in the stream.
    """

    def __init__(
        self,
        transcript_result_stream,
        request_id=None,
        language_code=None,
        media_sample_rate_hz=None,
        media_encoding=None,
        vocabulary_name=None,
        session_id=None,
        vocab_filter_name=None,
        vocab_filter_method=None,
        show_speaker_label=None,
        enable_channel_identification=None,
        number_of_channels=None,
    ):
        self.request_id: Optional[str] = request_id
        self.language_code: Optional[str] = language_code
        self.media_sample_rate_hz: Optional[int] = media_sample_rate_hz
        self.media_encoding: Optional[str] = media_encoding
        self.vocabulary_name: Optional[str] = vocabulary_name
        self.session_id: Optional[str] = session_id
        self.transcript_result_stream: TranscriptResultStream = transcript_result_stream
        self.vocab_filter_name: Optional[str] = vocab_filter_name
        self.vocab_filter_method: Optional[str] = vocab_filter_method
        self.show_speaker_label: Optional[bool] = show_speaker_label
        self.enable_channel_identification: Optional[
            bool
        ] = enable_channel_identification
        self.number_of_channels: Optional[int] = number_of_channels


class Transcript:
    """The transcription in a TranscriptEvent.

    :param results:
        Result objects that contain the results of transcribing a portion of the
        input audio stream. The array can be empty.
    """

    def __init__(self, results: List[Result]):
        self.results = results


class TranscriptEvent(BaseEvent):
    """Represents a set of transcription results from the server to the client.
    It contains one or more segments of the transcription.

    :param transcript:
        The transcription of the audio stream. The transcription is composed of
        all of the items in the results list.
    """

    def __init__(self, transcript: Transcript):
        self.transcript = transcript


class TranscriptResultStream(EventStream):
    """Transcription result stream containing returned TranscriptEvent output.

    Results are surfaced through the async iterator interface (i.e. async for)

    :raises BadRequestException:
        A client error occurred when the stream was created. Check the parameters
        of the request and try your request again.

    :raises LimitExceededException:
        Your client has exceeded one of the Amazon Transcribe limits, typically
        the limit on audio length. Break your audio stream into smaller chunks
        and try your request again.

    :raises InternalFailureException:
        A problem occurred while processing the audio.
        Amazon Transcribe terminated processing.

    :raises ConflictException:
        A new stream started with the same session ID.
        The current stream has been terminated.

    :raises ServiceUnavailableException:
        Service is currently unavailable. Try your request later.
    """


class StartStreamTranscriptionEventStream:
    """Event stream wrapper containing both input and output interfaces to
    Amazon Transcribe. This should only be created by the client.

    :param audio_stream:
        Audio input stream generated by client for new transcription requests.

    :param response: Response object from Amazon Transcribe.
    """

    def __init__(self, audio_stream: AudioStream, response):
        self._response = response
        self._audio_stream = audio_stream

    @property
    def response(self) -> StartStreamTranscriptionResponse:
        """Response object from Amazon Transcribe containing metadata and
        response output stream.
        """
        return self._response

    @property
    def input_stream(self) -> AudioStream:
        """Audio stream to Amazon Transcribe that takes input audio."""
        return self._audio_stream

    @property
    def output_stream(self) -> TranscriptResultStream:
        """Response stream containing transcribed event output."""
        return self.response.transcript_result_stream



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


from io import BufferedIOBase
from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    # We need to import this from _typeshed as this is not publicly exposed and
    # would otherwise require us to redefine this type to subclass
    # BufferedIOBase
    from _typeshed import ReadableBuffer


class BufferableByteStream(BufferedIOBase):
    """ BufferableByteStream will always be in non-blocking mode """

    def __init__(self):
        self._byte_chunks: list = []
        self.__done: bool = False
        self.__closed: bool = False

    def read(self, size=-1) -> Optional[bytes]:  # type: ignore
        if len(self._byte_chunks) < 1 and not self.__done:
            raise BlockingIOError("read")
        elif (self.__done and not self._byte_chunks) or self.closed:
            return b""

        temp_bytes = self._byte_chunks.pop(0)
        remaining_bytes = b""
        if size == -1:
            return temp_bytes
        elif size > 0:
            remaining_bytes = temp_bytes[size:]
            temp_bytes = temp_bytes[:size]
        else:
            remaining_bytes = temp_bytes
            temp_bytes = b""

        if len(remaining_bytes) > 0:
            self._byte_chunks.insert(0, remaining_bytes)
        return temp_bytes

    def read1(self, size=-1) -> Optional[bytes]:  # type: ignore
        return self.read(size)

    def readinto(self, b, read1=False):
        if not isinstance(b, memoryview):
            b = memoryview(b)
            b = b.cast("B")

        if read1:
            data = self.read1(len(b))
        else:
            data = self.read(len(b))

        if data is None:
            raise BlockingIOError("readinto")

        n = len(data)

        b[:n] = data

        return n

    def write(self, b: "ReadableBuffer") -> int:
        if not isinstance(b, bytes):
            type_ = type(b)
            raise ValueError(
                f"Unexpected value written to BufferableByteStream. "
                f"Only bytes are support but {type_} was provided."
            )

        if self.closed or self.__done:
            raise IOError("Stream is completed and doesn't support further writes.")

        if b:
            self._byte_chunks.append(b)

        return len(b)

    @property
    def closed(self) -> bool:
        return self.__closed

    def close(self):
        self._buffered_bytes_chunks = None
        self.__done = True
        self.__closed = True

    def end_stream(self):
        self.__done = True



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import json
from typing import Optional, Type, Any, List

import amazon_transcribe.exceptions as transcribe_exceptions
from amazon_transcribe.eventstream import BaseEvent
from amazon_transcribe.model import (
    StartStreamTranscriptionResponse,
    TranscriptResultStream,
    TranscriptEvent,
    Transcript,
    Result,
    Alternative,
    Item,
)
from amazon_transcribe.response import Response
from amazon_transcribe.exceptions import (
    ServiceException,
    BadRequestException,
    ConflictException,
    InternalFailureException,
    LimitExceededException,
    ServiceUnavailableException,
    UnknownServiceException,
    SerializationException,
)
from amazon_transcribe.utils import ensure_boolean


class TranscribeStreamingResponseParser:
    """Converts raw HTTP responses into modeled objects and exceptions.

    This class is not public and must not be consumed outside of this project.
    """

    def _get_error_code(self, http_response: Response) -> str:
        error_code = "Unknown"
        if "x-amzn-errortype" in http_response.headers:
            error_code = http_response.headers["x-amzn-errortype"]
            # Could be x-amzn-errortype: ValidationException:
            error_code = error_code.split(":")[0]
        return error_code

    def _get_error_message(self, body_bytes: bytes) -> str:
        error_message = "An unknown error was returned by the service"
        try:
            parsed_body = json.loads(body_bytes)
        except json.decoder.JSONDecodeError:
            return error_message
        if "Message" in parsed_body:
            error_message = parsed_body["Message"]
        elif "message" in parsed_body:
            error_message = parsed_body["message"]
        return error_message

    def parse_exception(
        self, http_response: Response, body_bytes: bytes
    ) -> ServiceException:
        error_code = self._get_error_code(http_response)
        error_message = self._get_error_message(body_bytes)
        if error_code == "BadRequestException":
            return BadRequestException(error_message)
        elif error_code == "ConflictException":
            return ConflictException(error_message)
        elif error_code == "InternalFailureException":
            return InternalFailureException(error_message)
        elif error_code == "LimitExceededException":
            return LimitExceededException(error_message)
        elif error_code == "ServiceUnavailableException":
            return ServiceUnavailableException(error_message)
        elif error_code == "SerializationException":
            return SerializationException(error_message)
        return UnknownServiceException(
            http_response.status_code, error_code, error_message,
        )

    def parse_start_stream_transcription_response(
        self, http_response: Response, body_stream: Any,
    ) -> StartStreamTranscriptionResponse:
        headers = http_response.headers
        request_id = headers.get("x-amzn-request-id")
        language_code = headers.get("x-amzn-transcribe-language-code")
        media_encoding = headers.get("x-amzn-transcribe-media-encoding")
        vocabulary_name = headers.get("x-amzn-transcribe-vocabulary-name")
        session_id = headers.get("x-amzn-transcribe-session-id")
        vocab_filter_name = headers.get("x-amzn-transcribe-vocabulary-filter-name")
        vocab_filter_method = headers.get("x-amzn-transcribe-vocabulary-filter-method")
        show_speaker_label = self._raw_value_to_bool(
            headers.get("x-amzn-transcribe-show-speaker-label")
        )
        enable_channel_identification = self._raw_value_to_bool(
            headers.get("x-amzn-transcribe-enable-channel-identification")
        )
        number_of_channels = self._raw_value_to_int(
            headers.get("x-amzn-transcribe-number-of-channels")
        )
        media_sample_rate_hz = self._raw_value_to_int(
            headers.get("x-amzn-transcribe-sample-rate")
        )

        transcript_result_stream = TranscriptResultStream(
            body_stream, TranscribeStreamingEventParser()
        )

        parsed_response = StartStreamTranscriptionResponse(
            transcript_result_stream=transcript_result_stream,
            request_id=request_id,
            language_code=language_code,
            media_sample_rate_hz=media_sample_rate_hz,
            media_encoding=media_encoding,
            vocabulary_name=vocabulary_name,
            session_id=session_id,
            vocab_filter_name=vocab_filter_name,
            vocab_filter_method=vocab_filter_method,
            show_speaker_label=show_speaker_label,
            enable_channel_identification=enable_channel_identification,
            number_of_channels=number_of_channels,
        )
        return parsed_response

    def _raw_value_to_int(self, value: Optional[str]) -> Optional[int]:
        if value:
            return int(value)
        return None

    def _raw_value_to_bool(self, value: Optional[str]) -> Optional[bool]:
        if value is not None:
            return ensure_boolean(value)
        return None


class TranscribeStreamingEventParser:
    def parse(self, raw_event) -> Optional[BaseEvent]:
        message_type = raw_event.headers.get(":message-type")
        if message_type in ["error", "exception"]:
            raise self._parse_event_exception(raw_event)
        elif message_type == "event":
            event_type = raw_event.headers.get(":event-type")
            raw_body = json.loads(raw_event.payload)
            if event_type == "TranscriptEvent":
                # TODO: Handle cases where the service returns an incorrect response
                return self._parse_transcript_event(raw_body)
        return None

    def _parse_transcript_event(self, current_node: Any) -> TranscriptEvent:
        transcript = self._parse_transcript(current_node.get("Transcript"))
        return TranscriptEvent(transcript)

    def _parse_transcript(self, current_node: Any) -> Transcript:
        results = self._parse_result_list(current_node.get("Results"))
        return Transcript(results)

    def _parse_result_list(self, current_node: Any) -> List[Result]:
        return [self._parse_result(e) for e in current_node]

    def _parse_result(self, current_node: Any) -> Result:
        alternatives = self._parse_alternative_list(current_node.get("Alternatives"))
        return Result(
            result_id=current_node.get("ResultId"),
            start_time=current_node.get("StartTime"),
            end_time=current_node.get("EndTime"),
            is_partial=current_node.get("IsPartial"),
            alternatives=alternatives,
            channel_id=current_node.get("ChannelId"),
        )

    def _parse_alternative_list(self, current_node: Any) -> List[Alternative]:
        return [self._parse_alternative(e) for e in current_node]

    def _parse_alternative(self, current_node: Any) -> Alternative:
        return Alternative(
            transcript=current_node.get("Transcript"),
            items=self._parse_item_list(current_node.get("Items")),
        )

    def _parse_item_list(self, current_node: Any) -> List[Item]:
        return [self._parse_item(e) for e in current_node]

    def _parse_item(self, current_node: Any) -> Item:
        return Item(
            start_time=current_node.get("StartTime"),
            end_time=current_node.get("EndTime"),
            item_type=current_node.get("Type"),
            content=current_node.get("Content"),
            vocabulary_filter_match=current_node.get("VocabularyFilterMatch"),
            speaker=current_node.get("Speaker"),
        )

    def _parse_event_exception(self, raw_event) -> ServiceException:
        exception_type: str = raw_event.headers.get(
            ":exception-type", "ServiceException"
        )
        exception_cls: Type[ServiceException] = getattr(
            transcribe_exceptions, exception_type, ServiceException
        )
        try:
            raw_body = json.loads(raw_event.payload)
        except ValueError:
            raw_body = {}
        exception_msg = raw_body.get("Message", "An unknown service exception occured")
        return exception_cls(exception_msg)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


class BaseEndpointResolver:
    """Asynchronous endpoint resolver"""

    async def resolve(self, region: str) -> str:
        """This is currently stubbed out until we have a complete
        and async endpoint resolution implementation.
        """
        raise NotImplementedError


class StaticEndpointResolver(BaseEndpointResolver):
    def __init__(self, endpoint):
        self._endpoint: str = endpoint

    async def resolve(self, region: str) -> str:
        """We ignore the region and return our static endpoint."""
        return self._endpoint


class _TranscribeRegionEndpointResolver(BaseEndpointResolver):
    async def resolve(self, region: str) -> str:
        """Apply region to transcribe uri template."""
        return f"https://transcribestreaming.{region}.amazonaws.com"



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


from typing import Any, Dict, Tuple, Optional

from amazon_transcribe.request import Request
from amazon_transcribe.structures import BufferableByteStream
from amazon_transcribe.utils import _add_required_headers
from amazon_transcribe.exceptions import SerializerException
from amazon_transcribe.model import (
    AudioEvent,
    StartStreamTranscriptionRequest,
    BaseEvent,
)


class TranscribeStreamingSerializer:
    """Convert StartStreamTranscriptionRequest into a
    Request object for streaming to the Transcribe service.
    """

    def _serialize_header(
        self, header: str, value: Any, prefix: str = "x-amzn-transcribe-",
    ) -> Dict[str, str]:
        if value is None:
            return {}
        else:
            return {f"{prefix}{header}": str(value)}

    def _serialize_str_header(
        self, header: str, value: Optional[str]
    ) -> Dict[str, str]:
        return self._serialize_header(header, value)

    def _serialize_int_header(
        self, header: str, value: Optional[int]
    ) -> Dict[str, str]:
        return self._serialize_header(header, value)

    def _serialize_bool_header(
        self, header: str, value: Optional[bool]
    ) -> Dict[str, str]:
        return self._serialize_header(header, value)

    def serialize_start_stream_transcription_request(
        self, endpoint: str, request_shape: StartStreamTranscriptionRequest
    ) -> Request:

        method = "POST"
        request_uri = "/stream-transcription"

        headers: Dict[str, str] = {}
        headers.update(
            self._serialize_str_header("language-code", request_shape.language_code)
        )
        headers.update(
            self._serialize_int_header(
                "sample-rate", request_shape.media_sample_rate_hz
            )
        )
        headers.update(
            self._serialize_str_header("media-encoding", request_shape.media_encoding)
        )
        headers.update(
            self._serialize_str_header("vocabulary-name", request_shape.vocabulary_name)
        )
        headers.update(
            self._serialize_str_header("session-id", request_shape.session_id)
        )
        headers.update(
            self._serialize_str_header(
                "vocabulary-filter-method", request_shape.vocab_filter_method,
            )
        )
        headers.update(
            self._serialize_str_header(
                "vocabulary-filter-name", request_shape.vocab_filter_name,
            )
        )
        headers.update(
            self._serialize_bool_header(
                "show-speaker-label", request_shape.show_speaker_label,
            )
        )
        headers.update(
            self._serialize_bool_header(
                "enable-channel-identification",
                request_shape.enable_channel_identification,
            )
        )
        headers.update(
            self._serialize_int_header(
                "number-of-channels", request_shape.number_of_channels,
            )
        )

        _add_required_headers(endpoint, headers)

        body = BufferableByteStream()

        request = Request(
            endpoint=endpoint,
            path=request_uri,
            method=method,
            headers=headers,
            body=body,
        )
        return request


SERIALIZED_EVENT = Tuple[Dict, bytes]


class EventSerializer:
    def serialize(self, audio_event: BaseEvent) -> SERIALIZED_EVENT:
        raise NotImplementedError("serialize")


class AudioEventSerializer(EventSerializer):
    """Convert AudioEvent objects into payload and header outputs for eventstreams"""

    def serialize(self, event: BaseEvent) -> SERIALIZED_EVENT:
        if isinstance(event, AudioEvent):
            return self._serialize_audio_event(event)
        raise SerializerException(f'Unexpected event type encountered: "{type(event)}"')

    def _serialize_audio_event(self, audio_event: AudioEvent):
        headers = {
            ":message-type": "event",
            ":event-type": "AudioEvent",
            ":content-type": "application/octet-stream",
        }
        return headers, audio_event.payload



#!/usr/bin/env python
import codecs
import os.path
import re

from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))


def read(*parts):
    return codecs.open(os.path.join(here, *parts), "r", encoding="utf-8").read()


def find_version(*file_paths):
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]", version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


requires = ["awscrt==0.11.15"]

setup(
    name="amazon-transcribe",
    version=find_version("amazon_transcribe", "__init__.py"),
    description="Async Python SDK for Amazon Transcribe Streaming",
    long_description=open("README.md", "r", encoding="utf-8").read(),
    long_description_content_type='text/markdown',
    author="Amazon Web Services",
    url="https://github.com/awslabs/amazon-transcribe-streaming-sdk",
    scripts=[],
    packages=find_packages(exclude=["tests*"]),
    include_package_data=True,
    install_requires=requires,
    extras_require={},
    license="Apache License 2.0",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Natural Language :: English",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


import asyncio

from io import BytesIO
from threading import Lock
from urllib.parse import urlparse, ParseResult
from awscrt import io, http
from concurrent.futures import Future
from typing import Optional, List, Tuple, Union, Awaitable, AsyncGenerator

from amazon_transcribe.exceptions import HTTPException
from amazon_transcribe.response import Response

HeadersList = List[Tuple[str, str]]


class AwsCrtHttpResponse:
    def __init__(self):
        self._stream = None
        self._status_code_future: Future[int] = Future()
        self._headers_future: Future[HeadersList] = Future()
        self._chunk_futures: List[Future[bytes]] = []
        self._received_chunks: List[bytes] = []
        self._chunk_lock: Lock = Lock()

    async def resolve_response(self) -> Response:
        return Response(
            status_code=await self.status_code, headers=dict(await self.headers),
        )

    async def consume_body(self) -> bytes:
        body = b""
        async for chunk in self.chunks():
            body += chunk
        return body

    def _set_stream(self, stream: http.HttpClientStream):
        if self._stream is not None:
            raise HTTPException("Stream already set on AwsCrtHttpResponse object")
        self._stream = stream
        self._stream.completion_future.add_done_callback(self._on_complete)
        self._stream.activate()

    @property
    def status_code(self) -> Awaitable[int]:
        return asyncio.wrap_future(self._status_code_future)

    @property
    def headers(self) -> Awaitable[HeadersList]:
        return asyncio.wrap_future(self._headers_future)

    @property
    def done(self) -> Awaitable[bool]:
        return asyncio.wrap_future(self._stream.completion_future)

    def get_chunk(self) -> Awaitable[bytes]:
        with self._chunk_lock:
            future: Future[bytes] = Future()
            # TODO: update backpressure window
            if self._received_chunks:
                chunk = self._received_chunks.pop(0)
                future.set_result(chunk)
            elif self._stream.completion_future.done():
                future.set_result(b"")
            else:
                self._chunk_futures.append(future)
            return asyncio.wrap_future(future)

    async def chunks(self) -> AsyncGenerator[bytes, None]:
        while True:
            chunk = await self.get_chunk()
            if chunk:
                yield chunk
            else:
                break

    def _on_headers(self, status_code: int, headers: HeadersList, **kwargs):
        self._status_code_future.set_result(status_code)
        self._headers_future.set_result(headers)

    def _on_body(self, chunk: bytes, **kwargs):
        with self._chunk_lock:
            # TODO: update back pressure window
            if self._chunk_futures:
                future = self._chunk_futures.pop(0)
                future.set_result(chunk)
            else:
                self._received_chunks.append(chunk)

    def _on_complete(self, completion_future):
        with self._chunk_lock:
            if self._chunk_futures:
                future = self._chunk_futures.pop(0)
                future.set_result(b"")


class AwsCrtHttpSessionManager:
    _HTTP_PORT = 80
    _HTTPS_PORT = 443
    HTTP_CONNECTION_CLS = http.HttpClientConnection

    def __init__(self, eventloop):
        # TODO: accept an AWSEventLoop object or similar
        self._client_bootstrap = eventloop
        self._tls_ctx = io.ClientTlsContext(io.TlsContextOptions())
        self._socket_options = io.SocketOptions()
        self._connections = {}

    async def _create_connection(
        self, parsed_url: ParseResult
    ) -> http.HttpClientConnection:
        if parsed_url.scheme == "http":
            port = self._HTTP_PORT
            tls_connection_options = None
        else:
            port = self._HTTPS_PORT
            tls_connection_options = self._tls_ctx.new_connection_options()
            tls_connection_options.set_server_name(parsed_url.hostname)
            tls_connection_options.set_alpn_list(["h2"])
        if parsed_url.port is not None:
            port = parsed_url.port
        connect_future = self.HTTP_CONNECTION_CLS.new(
            bootstrap=self._client_bootstrap,
            host_name=parsed_url.hostname,
            port=port,
            socket_options=self._socket_options,
            tls_connection_options=tls_connection_options,
        )
        connection = await asyncio.wrap_future(connect_future)
        if connection.version is not http.HttpVersion.Http2:
            connection.close()
            raise HTTPException(
                "HTTP/2 could not be negotiated: %s" % connection.version
            )
        return connection

    async def _get_connection(
        self, parsed_url: ParseResult,
    ) -> http.HttpClientConnection:
        # TODO: Use CRT connection pooling instead of this basic kind
        if not parsed_url.hostname:
            raise HTTPException("Invalid host name: %s" % parsed_url.hostname)
        connection_key = (
            parsed_url.scheme,
            parsed_url.hostname,
            parsed_url.port,
        )
        if connection_key in self._connections:
            return self._connections[connection_key]
        else:
            connection = await self._create_connection(parsed_url)
            self._connections[connection_key] = connection
            return connection

    def _get_path(self, parsed_url: ParseResult) -> str:
        path = parsed_url.path
        if not path:
            path = "/"
        if parsed_url.query:
            path = path + "?" + parsed_url.query
        return path

    async def make_request(
        self,
        url: str,
        method: str = "GET",
        headers: Optional[HeadersList] = None,
        body: Union[bytes, BytesIO, None] = None,
    ) -> AwsCrtHttpResponse:
        if isinstance(headers, list):
            headers = http.HttpHeaders(headers)
        if isinstance(body, bytes):
            body = BytesIO(body)

        parsed_url = urlparse(url)
        request = http.HttpRequest(
            method=method,
            path=self._get_path(parsed_url),
            headers=headers,
            body_stream=body,
        )

        connection = await self._get_connection(parsed_url)
        response = AwsCrtHttpResponse()
        stream = connection.request(request, response._on_headers, response._on_body,)
        response._set_stream(stream)
        return response



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


from amazon_transcribe.model import TranscriptEvent, TranscriptResultStream


class TranscriptResultStreamHandler:
    def __init__(self, transcript_result_stream: TranscriptResultStream):
        self._transcript_result_stream = transcript_result_stream

    async def handle_events(self):
        """Process generic incoming events from Amazon Transcribe
        and delegate to appropriate sub-handlers.
        """
        async for event in self._transcript_result_stream:
            if isinstance(event, TranscriptEvent):
                await self.handle_transcript_event(event)

    async def handle_transcript_event(self, transcript_event: TranscriptEvent):
        """Specific handling for TranscriptionEvent responses from Amazon Transcribe.

        This should be implemented by the end user with desired data handling.
        """
        raise NotImplementedError



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.


__version__ = "0.4.0"

from awscrt.io import ClientBootstrap, DefaultHostResolver, EventLoopGroup


class AWSCRTEventLoop:
    def __init__(self):
        self.bootstrap = self._initialize_default_loop()

    def _initialize_default_loop(self):
        event_loop_group = EventLoopGroup(1)
        host_resolver = DefaultHostResolver(event_loop_group)
        return ClientBootstrap(event_loop_group, host_resolver,)



# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Binary Event Stream Decoding """
import uuid
import datetime
from typing import Any, Callable, Dict, AsyncGenerator, Optional, Tuple, Union, Type

from binascii import crc32
from struct import unpack, pack

from amazon_transcribe.structures import BufferableByteStream


# TODO: move this to another module when we fix structure
# These are needed for event signing
import hmac
from hashlib import sha256
from binascii import hexlify
from amazon_transcribe.auth import Credentials


# byte length of the prelude (total_length + header_length + prelude_crc)
_PRELUDE_LENGTH = 12
_MAX_HEADERS_LENGTH = 128 * 1024  # 128 Kb
_MAX_HEADER_VALUE_BYTE_LENGTH = 32 * 1024 - 1
_MAX_PAYLOAD_LENGTH = 16 * 1024 ** 2  # 16 Mb

HEADER_VALUE = Union[bool, bytes, int, str]


class ParserError(Exception):
    """Base binary flow encoding parsing exception."""


class DuplicateHeader(ParserError):
    """Duplicate header found in the event."""

    def __init__(self, header: str):
        message = 'Duplicate header present: "%s"' % header
        super(DuplicateHeader, self).__init__(message)


class InvalidHeadersLength(ParserError):
    """Headers length is longer than the maximum."""

    def __init__(self, length: int):
        message = "Header length of %s exceeded the maximum of %s" % (
            length,
            _MAX_HEADERS_LENGTH,
        )
        super().__init__(message)


class InvalidPayloadLength(ParserError):
    """Payload length is longer than the maximum."""

    def __init__(self, length: int):
        message = "Payload length of %s exceeded the maximum of %s" % (
            length,
            _MAX_PAYLOAD_LENGTH,
        )
        super().__init__(message)


class ChecksumMismatch(ParserError):
    """Calculated checksum did not match the expected checksum."""

    def __init__(self, expected: int, calculated: int):
        message = "Checksum mismatch: expected 0x%08x, calculated 0x%08x" % (
            expected,
            calculated,
        )
        super().__init__(message)


class SerializationError(Exception):
    """Base binary flow encoding serialization exception."""


class InvalidHeaderValue(SerializationError):
    def __init__(self, value):
        message = f"Invalid header value type: {type(value)}"
        super(InvalidHeaderValue, self).__init__(message)
        self.value = value


class HeaderBytesExceedMaxLength(SerializationError):
    def __init__(self, length):
        message = (
            f"Headers exceeded max serialization "
            f"length of 128 KiB at {length} bytes"
        )
        super(HeaderBytesExceedMaxLength, self).__init__(message)


class HeaderValueBytesExceedMaxLength(SerializationError):
    def __init__(self, length):
        message = (
            f"Header bytes value exceeds max serialization "
            f"length of (32 KiB - 1) at {length} bytes"
        )
        super(HeaderValueBytesExceedMaxLength, self).__init__(message)


class PayloadBytesExceedMaxLength(SerializationError):
    def __init__(self, length):
        message = (
            f"Payload exceeded max serialization " f"length of 16 MiB at {length} bytes"
        )
        super(PayloadBytesExceedMaxLength, self).__init__(message)


class HeaderValue:
    """A wrapper class for explicit header serialization."""

    def __init__(self, *args, **kwargs):
        raise NotImplementedError


class Int8HeaderValue(HeaderValue):
    """Value that should be explicitly serialized as an int8."""

    def __init__(self, value: int):
        self.value = value


class Int16HeaderValue(HeaderValue):
    """Value that should be explicitly serialized as an int16"""

    def __init__(self, value: int):
        self.value = value


class Int32HeaderValue(HeaderValue):
    """Value that should be explicitly serialized as an int32"""

    def __init__(self, value: int):
        self.value = value


class Int64HeaderValue(HeaderValue):
    """Value that should be explicitly serialized as an int64"""

    def __init__(self, value: int):
        self.value = value


# Possible types for serializing headers differs from possible types returned when decoding
HEADER_SERIALIZATION_VALUE = Union[
    bool, bytes, int, str, uuid.UUID, datetime.datetime, HeaderValue
]
HEADERS_SERIALIZATION_DICT = Dict[str, HEADER_SERIALIZATION_VALUE]


class EventStreamMessageSerializer:
    DEFAULT_INT_TYPE: Type[HeaderValue] = Int32HeaderValue

    def serialize(self, headers: HEADERS_SERIALIZATION_DICT, payload: bytes) -> bytes:
        # TODO: Investigate preformance of this once we can make requests
        if len(payload) > _MAX_PAYLOAD_LENGTH:
            raise PayloadBytesExceedMaxLength(len(payload))
        # The encoded headers are variable length and this length
        # is required to generate the prelude, generate the headers first
        encoded_headers = self.encode_headers(headers)
        if len(encoded_headers) > _MAX_HEADERS_LENGTH:
            raise HeaderBytesExceedMaxLength(len(encoded_headers))
        prelude_bytes = self._encode_prelude(encoded_headers, payload)
        # Calculate the prelude_crc and it's byte representation
        prelude_crc = self._calculate_checksum(prelude_bytes)
        prelude_crc_bytes = pack("!I", prelude_crc)
        messages_bytes = prelude_crc_bytes + encoded_headers + payload
        # Calculate the checksum continuing from the prelude crc
        final_crc = self._calculate_checksum(messages_bytes, crc=prelude_crc)
        final_crc_bytes = pack("!I", final_crc)
        return prelude_bytes + messages_bytes + final_crc_bytes

    def encode_headers(self, headers: HEADERS_SERIALIZATION_DICT) -> bytes:
        encoded = b""
        for key, val in headers.items():
            encoded += self._encode_header_key(key)
            encoded += self._encode_header_val(val)
        return encoded

    def _encode_header_key(self, key: str) -> bytes:
        enc = key.encode("utf-8")
        return pack("B", len(enc)) + enc

    def _encode_header_val(self, val: HEADER_SERIALIZATION_VALUE) -> bytes:
        # Handle booleans first to avoid being viewed as ints
        if val is True:
            return b"\x00"
        elif val is False:
            return b"\x01"

        if isinstance(val, int):
            val = self.DEFAULT_INT_TYPE(val)

        if isinstance(val, Int8HeaderValue):
            return b"\x02" + pack("!b", val.value)
        elif isinstance(val, Int16HeaderValue):
            return b"\x03" + pack("!h", val.value)
        elif isinstance(val, Int32HeaderValue):
            return b"\x04" + pack("!i", val.value)
        elif isinstance(val, Int64HeaderValue):
            return b"\x05" + pack("!q", val.value)
        elif isinstance(val, bytes):
            # Byte arrays are prefaced with a 16bit length, but are restricted
            # to a max length of 2**15 - 1, enforce this explicitly
            if len(val) > _MAX_HEADER_VALUE_BYTE_LENGTH:
                raise HeaderValueBytesExceedMaxLength(len(val))
            return b"\x06" + pack("!H", len(val)) + val
        elif isinstance(val, str):
            utf8_string = val.encode("utf-8")
            # Strings are prefaced with a 16bit length, but are restricted
            # to a max length of 2**15 - 1, enforce this explicitly
            if len(utf8_string) > _MAX_HEADER_VALUE_BYTE_LENGTH:
                raise HeaderValueBytesExceedMaxLength(len(utf8_string))
            return b"\x07" + pack("!H", len(utf8_string)) + utf8_string
        elif isinstance(val, datetime.datetime):
            ms_timestamp = int(val.timestamp() * 1000)
            return b"\x08" + pack("!q", ms_timestamp)
        elif isinstance(val, uuid.UUID):
            return b"\x09" + val.bytes
        raise InvalidHeaderValue(val)

    def _encode_prelude(self, encoded_headers: bytes, payload: bytes) -> bytes:
        header_length = len(encoded_headers)
        payload_length = len(payload)
        total_length = header_length + payload_length + 16
        return pack("!II", total_length, header_length)

    def _calculate_checksum(self, data: bytes, crc: int = 0) -> int:
        return crc32(data, crc) & 0xFFFFFFFF


class BaseEvent:
    """Base class for typed events sent over event stream with service.

    :param payload: bytes payload to be sent with event
    :param event_payload: boolean stating if event has a payload
    """

    def __init__(self, payload: bytes, event_payload: Optional[bool] = None):
        self.payload = payload
        self.event_payload = event_payload
        self.event = True


class BaseStream:
    """Base class for EventStream established between client and Transcribe Service.

    These streams will always be established automatically by the client.
    """

    def __init__(
        self,
        input_stream=None,
        event_serializer=None,
        eventstream_serializer=None,
        event_signer=None,
        initial_signature=None,
        credential_resolver=None,
    ):
        if input_stream is None:
            input_stream = BufferableByteStream()
        self._input_stream: BufferableByteStream = input_stream
        # TODO: Cant type due to circular import
        self._event_serializer = event_serializer
        if eventstream_serializer is None:
            eventstream_serializer = EventStreamMessageSerializer()
        self._eventstream_serializer = eventstream_serializer
        self._event_signer = event_signer
        self._prior_signature = initial_signature
        self._credential_resolver = credential_resolver

    async def send_event(self, event: BaseEvent):
        headers, payload = self._event_serializer.serialize(event)
        event_bytes = self._eventstream_serializer.serialize(headers, payload)
        signed_bytes = await self._sign_event(event_bytes)
        self._input_stream.write(signed_bytes)

    async def end_stream(self):
        signed_bytes = await self._sign_event(b"")
        self._input_stream.write(signed_bytes)
        self._input_stream.end_stream()

    async def _sign_event(self, event_bytes):
        creds = await self._credential_resolver.get_credentials()
        signed_headers = self._event_signer.sign(
            event_bytes, self._prior_signature, creds
        )
        self._prior_signature = signed_headers.get(":chunk-signature")
        return self._eventstream_serializer.serialize(signed_headers, event_bytes)


class DecodeUtils:
    """Unpacking utility functions used in the decoder.

    All methods on this class take raw bytes and return  a tuple containing
    the value parsed from the bytes and the number of bytes consumed to parse
    that value.
    """

    UINT8_BYTE_FORMAT = "!B"
    UINT16_BYTE_FORMAT = "!H"
    UINT32_BYTE_FORMAT = "!I"
    INT8_BYTE_FORMAT = "!b"
    INT16_BYTE_FORMAT = "!h"
    INT32_BYTE_FORMAT = "!i"
    INT64_BYTE_FORMAT = "!q"
    PRELUDE_BYTE_FORMAT = "!III"

    # uint byte size to unpack format
    UINT_BYTE_FORMAT = {
        1: UINT8_BYTE_FORMAT,
        2: UINT16_BYTE_FORMAT,
        4: UINT32_BYTE_FORMAT,
    }

    @staticmethod
    def unpack_true(data: bytes) -> Tuple[bool, int]:
        """This method consumes none of the provided bytes and returns True"""
        return True, 0

    @staticmethod
    def unpack_false(data: bytes) -> Tuple[bool, int]:
        """This method consumes none of the provided bytes and returns False."""
        return False, 0

    @staticmethod
    def unpack_uint8(data: bytes) -> Tuple[int, int]:
        """Parse an unsigned 8-bit integer from the bytes."""
        value = unpack(DecodeUtils.UINT8_BYTE_FORMAT, data[:1])[0]
        return value, 1

    @staticmethod
    def unpack_uint32(data: bytes) -> Tuple[int, int]:
        """Parse an unsigned 32-bit integer from the bytes."""
        value = unpack(DecodeUtils.UINT32_BYTE_FORMAT, data[:4])[0]
        return value, 4

    @staticmethod
    def unpack_int8(data):
        """Parse a signed 8-bit integer from the bytes.

        :type data: bytes
        :param data: The bytes to parse from.

        :rtype: (int, int)
        :returns: A tuple containing the (parsed integer value, bytes consumed)
        """
        value = unpack(DecodeUtils.INT8_BYTE_FORMAT, data[:1])[0]
        return value, 1

    @staticmethod
    def unpack_int16(data: bytes) -> Tuple[int, int]:
        """Parse a signed 16-bit integer from the bytes."""
        value = unpack(DecodeUtils.INT16_BYTE_FORMAT, data[:2])[0]
        return value, 2

    @staticmethod
    def unpack_int32(data: bytes) -> Tuple[int, int]:
        """Parse a signed 32-bit integer from the bytes."""
        value = unpack(DecodeUtils.INT32_BYTE_FORMAT, data[:4])[0]
        return value, 4

    @staticmethod
    def unpack_int64(data: bytes) -> Tuple[int, int]:
        """Parse a signed 64-bit integer from the bytes."""
        value = unpack(DecodeUtils.INT64_BYTE_FORMAT, data[:8])[0]
        return value, 8

    @staticmethod
    def unpack_byte_array(data: bytes, length_byte_size=2) -> Tuple[bytes, int]:
        """Parse a variable length byte array from the bytes.

        The bytes are expected to be in the following format:
            [ length ][0 ... length bytes]
        where length is an unsigned integer represented in the smallest number
        of bytes to hold the maximum length of the array.
        """
        uint_byte_format = DecodeUtils.UINT_BYTE_FORMAT[length_byte_size]
        length = unpack(uint_byte_format, data[:length_byte_size])[0]
        bytes_end = length + length_byte_size
        array_bytes = data[length_byte_size:bytes_end]
        return array_bytes, bytes_end

    @staticmethod
    def unpack_utf8_string(data: bytes, length_byte_size=2) -> Tuple[str, int]:
        """Parse a variable length utf-8 string from the bytes.

        The bytes are expected to be in the following format:
            [ length ][0 ... length bytes]
        where length is an unsigned integer represented in the smallest number
        of bytes to hold the maximum length of the array and the following
        bytes are a valid utf-8 string.
        """
        array_bytes, consumed = DecodeUtils.unpack_byte_array(data, length_byte_size)
        return array_bytes.decode("utf-8"), consumed

    @staticmethod
    def unpack_uuid(data: bytes) -> Tuple[bytes, int]:
        """Parse a 16-byte uuid from the bytes."""
        return data[:16], 16

    @staticmethod
    def unpack_prelude(data: bytes) -> Tuple[Tuple[Any, ...], int]:
        """Parse the prelude for an event stream message from the bytes.

        The prelude for an event stream message has the following format:
            [total_length][header_length][prelude_crc]
        where each field is an unsigned 32-bit integer.
        """
        return (unpack(DecodeUtils.PRELUDE_BYTE_FORMAT, data), _PRELUDE_LENGTH)


def _validate_checksum(data: bytes, checksum: int, crc=0):
    # To generate the same numeric value across all Python versions and
    # platforms use crc32(data) & 0xffffffff.
    computed_checksum = crc32(data, crc) & 0xFFFFFFFF
    if checksum != computed_checksum:
        raise ChecksumMismatch(checksum, computed_checksum)


class MessagePrelude:
    """Represents the prelude of an event stream message. """

    def __init__(self, total_length: int, headers_length: int, crc: int):
        self.total_length = total_length
        self.headers_length = headers_length
        self.crc = crc

    @property
    def payload_length(self) -> int:
        """Calculates the total payload length.

        The extra minus 4 bytes is for the message CRC.
        """
        return self.total_length - self.headers_length - _PRELUDE_LENGTH - 4

    @property
    def payload_end(self) -> int:
        """Calculates the byte offset for the end of the message payload.

        The extra minus 4 bytes is for the message CRC.
        """
        return self.total_length - 4

    @property
    def headers_end(self) -> int:
        """Calculates the byte offset for the end of the message headers."""
        return _PRELUDE_LENGTH + self.headers_length


class EventStreamMessage:
    """Represents an event stream message. """

    def __init__(self, prelude, headers, payload, crc):
        self.prelude: MessagePrelude = prelude
        self.headers: Dict = headers
        self.payload: BufferableByteStream = payload
        self.crc: int = crc

    def to_response_dict(self, status_code=200) -> Dict[str, Any]:
        message_type = self.headers.get(":message-type")
        if message_type == "error" or message_type == "exception":
            status_code = 400
        return {
            "status_code": status_code,
            "headers": self.headers,
            "body": self.payload,
        }


class EventStreamHeaderParser:
    """ Parses the event headers from an event stream message.

    Expects all of the header data upfront and creates a dictionary of headers
    to return. This object can be reused multiple times to parse the headers
    from multiple event stream messages.
    """

    # Maps header type to appropriate unpacking function
    # These unpacking functions return the value and the amount unpacked
    _HEADER_TYPE_MAP: Dict[int, Callable[[bytes], Tuple[HEADER_VALUE, int]]] = {
        # boolean_true
        0: DecodeUtils.unpack_true,
        # boolean_false
        1: DecodeUtils.unpack_false,
        # byte
        2: DecodeUtils.unpack_int8,
        # short
        3: DecodeUtils.unpack_int16,
        # integer
        4: DecodeUtils.unpack_int32,
        # long
        5: DecodeUtils.unpack_int64,
        # byte_array
        6: DecodeUtils.unpack_byte_array,
        # string
        7: DecodeUtils.unpack_utf8_string,
        # timestamp
        8: DecodeUtils.unpack_int64,
        # uuid
        9: DecodeUtils.unpack_uuid,
    }

    def __init__(self):
        self._data = None

    def parse(self, data: bytes) -> Dict[str, HEADER_VALUE]:
        """Parses the event stream headers from an event stream message."""
        self._data = data
        return self._parse_headers()

    def _parse_headers(self) -> Dict[str, HEADER_VALUE]:
        headers = {}
        while self._data:
            name, value = self._parse_header()
            if name in headers:
                raise DuplicateHeader(name)
            headers[name] = value
        return headers

    def _parse_header(self) -> Tuple[str, HEADER_VALUE]:
        name = self._parse_name()
        value = self._parse_value()
        return name, value

    def _parse_name(self) -> str:
        name, consumed = DecodeUtils.unpack_utf8_string(self._data, 1)
        self._advance_data(consumed)
        return name

    def _parse_type(self) -> int:
        type, consumed = DecodeUtils.unpack_uint8(self._data)
        self._advance_data(consumed)
        return type

    def _parse_value(self) -> HEADER_VALUE:
        header_type = self._parse_type()
        value_unpacker = self._HEADER_TYPE_MAP[header_type]
        value, consumed = value_unpacker(self._data)
        self._advance_data(consumed)
        return value

    def _advance_data(self, consumed):
        self._data = self._data[consumed:]


class EventStreamBuffer:
    """Streaming based event stream buffer

    A buffer class that wraps bytes from an event stream providing parsed
    messages as they become available via an iterable interface.
    """

    def __init__(self):
        self._data: bytes = b""
        self._prelude = None
        self._header_parser = EventStreamHeaderParser()

    def add_data(self, data: bytes):
        """Add data to the buffer."""
        self._data += data

    def _validate_prelude(self, prelude: MessagePrelude):
        if prelude.headers_length > _MAX_HEADERS_LENGTH:
            raise InvalidHeadersLength(prelude.headers_length)

        if prelude.payload_length > _MAX_PAYLOAD_LENGTH:
            raise InvalidPayloadLength(prelude.payload_length)

    def _parse_prelude(self) -> MessagePrelude:
        prelude_bytes = self._data[:_PRELUDE_LENGTH]
        raw_prelude, _ = DecodeUtils.unpack_prelude(prelude_bytes)
        prelude = MessagePrelude(*raw_prelude)
        self._validate_prelude(prelude)
        # The minus 4 removes the prelude crc from the bytes to be checked
        _validate_checksum(prelude_bytes[: _PRELUDE_LENGTH - 4], prelude.crc)
        return prelude

    def _parse_headers(self) -> Dict[str, str]:
        header_bytes = self._data[_PRELUDE_LENGTH : self._prelude.headers_end]
        return self._header_parser.parse(header_bytes)

    def _parse_payload(self) -> bytes:
        prelude = self._prelude
        payload_bytes = self._data[prelude.headers_end : prelude.payload_end]
        return payload_bytes

    def _parse_message_crc(self) -> int:
        prelude = self._prelude
        crc_bytes = self._data[prelude.payload_end : prelude.total_length]
        message_crc, _ = DecodeUtils.unpack_uint32(crc_bytes)
        return message_crc

    def _parse_message_bytes(self) -> bytes:
        # The minus 4 includes the prelude crc to the bytes to be checked
        message_bytes = self._data[_PRELUDE_LENGTH - 4 : self._prelude.payload_end]
        return message_bytes

    def _validate_message_crc(self) -> int:
        message_crc = self._parse_message_crc()
        message_bytes = self._parse_message_bytes()
        _validate_checksum(message_bytes, message_crc, crc=self._prelude.crc)
        return message_crc

    def _parse_message(self) -> EventStreamMessage:
        crc = self._validate_message_crc()
        headers = self._parse_headers()
        payload = self._parse_payload()
        message = EventStreamMessage(self._prelude, headers, payload, crc)
        self._prepare_for_next_message()
        return message

    def _prepare_for_next_message(self):
        # Advance the data and reset the current prelude
        self._data = self._data[self._prelude.total_length :]
        self._prelude = None

    def next(self) -> EventStreamMessage:
        """Provides the next available message parsed from the stream"""
        if len(self._data) < _PRELUDE_LENGTH:
            raise StopIteration()

        if self._prelude is None:
            self._prelude = self._parse_prelude()

        if len(self._data) < self._prelude.total_length:
            raise StopIteration()

        return self._parse_message()

    def __next__(self):
        return self.next()

    def __iter__(self):
        return self


class EventStream:
    """Wrapper class for an event stream body.

    This wraps the underlying streaming body, parsing it for individual events
    and yielding them as they come available through the async iterator interface.
    """

    def __init__(self, raw_stream, parser):
        self._raw_stream = raw_stream
        self._parser = parser
        self._event_generator: AsyncGenerator = self._create_raw_event_generator()

    async def __aiter__(self):
        async for event in self._event_generator:
            parsed_event = self._parser.parse(event)
            yield parsed_event

    async def _create_raw_event_generator(self) -> AsyncGenerator:
        event_stream_buffer = EventStreamBuffer()
        async for chunk in self._raw_stream.chunks():
            event_stream_buffer.add_data(chunk)
            for event in event_stream_buffer:
                yield event


def _utc_now() -> datetime.datetime:
    return datetime.datetime.now(datetime.timezone.utc)


class EventSigner:
    _ISO8601_TIMESTAMP_FMT = "%Y%m%dT%H%M%SZ"
    _NOW_TYPE = Optional[Callable[[], datetime.datetime]]

    def __init__(
        self, signing_name: str, region: str, utc_now: _NOW_TYPE = None,
    ):
        self.signing_name = signing_name
        self.region = region
        self.serializer = EventStreamMessageSerializer()
        if utc_now is None:
            utc_now = _utc_now
        self._utc_now = utc_now

    def sign(
        self, payload: bytes, prior_signature: bytes, credentials: Credentials
    ) -> HEADERS_SERIALIZATION_DICT:
        now = self._utc_now()
        headers: HEADERS_SERIALIZATION_DICT = {
            ":date": now,
        }
        timestamp = now.strftime(self._ISO8601_TIMESTAMP_FMT)
        string_to_sign = self._string_to_sign(
            timestamp, headers, payload, prior_signature
        )
        event_signature = self._sign_event(timestamp, string_to_sign, credentials)
        headers[":chunk-signature"] = event_signature
        return headers

    def _keypath(self, timestamp: str) -> str:
        parts = [
            timestamp[:8],  # Only using the YYYYMMDD
            self.region,
            self.signing_name,
            "aws4_request",
        ]
        return "/".join(parts)

    def _string_to_sign(
        self,
        timestamp: str,
        headers: HEADERS_SERIALIZATION_DICT,
        payload: bytes,
        prior_signature: bytes,
    ) -> str:
        encoded_headers = self.serializer.encode_headers(headers)
        parts = [
            "AWS4-HMAC-SHA256-PAYLOAD",
            timestamp,
            self._keypath(timestamp),
            hexlify(prior_signature).decode("utf-8"),
            sha256(encoded_headers).hexdigest(),
            sha256(payload).hexdigest(),
        ]
        return "\n".join(parts)

    def _hmac(self, key: bytes, msg: bytes) -> bytes:
        return hmac.new(key, msg, sha256).digest()

    def _sign_event(
        self, timestamp: str, string_to_sign: str, credentials: Credentials
    ) -> bytes:
        key = credentials.secret_access_key.encode("utf-8")
        today = timestamp[:8].encode("utf-8")  # Only using the YYYYMMDD
        k_date = self._hmac(b"AWS4" + key, today)
        k_region = self._hmac(k_date, self.region.encode("utf-8"))
        k_service = self._hmac(k_region, self.signing_name.encode("utf-8"))
        k_signing = self._hmac(k_service, b"aws4_request")
        return self._hmac(k_signing, string_to_sign.encode("utf-8"))
